#include "hime.h"
#include "pho-kbm-name.h"

struct PHO_KBM_NAME kbm_sel[]= {
 {N_("標準(零壹)"), "zo"},
 {N_("倚天"), "et"},
 {N_("倚天 26 鍵"), "et26"},
 {N_("許氏(國音,自然)"), "hsu"},
 {N_("拼音"), "pinyin"},
 {N_("拼音無聲調"), "pinyin-no-tone"},
 {N_("Dvorak"), "dvorak"},
 {N_("IBM"), "ibm"},
 {N_("神通"), "mitac"},
 {N_("Colemak"), "colemak"},
 {NULL, NULL}
};
